/*
 * asn_config.h - configures the ANSI/non ansi, defines 
 *                decoder alloc routines and buffer routines
 *
 * MS 91 
 * Copyright (C) 1992 Michael Sample and the University of British Columbia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 */

#ifndef _asn_config_h_
#define _asn_config_h_

#include <stdio.h>
#include <ctype.h> /*  for isprint() in asn_octs.c */
#include <setjmp.h> /* for jmp_buf type, setjmp and longjmp */

/* for pow() used in asn_real.c - must include to avoid casting err on pow */
#include <math.h>  

/*
 * define IEEE_REAL_FMT if your system/compiler uses the native ieee double
 * this should improve the performance of encoding reals.
 * If your system has the IEEE library routines (iszero, isinf etc)
 * then define IEEE_REAL_LIB.  If neither are defined then
 * frexp is used.  Performance is probaby best for IEEE_REAL_FMT.
 *
 *  #define IEEE_REAL_FMT
 *  #define IEEE_REAL_LIB
 */


/* used to test if optionals are present */
#define NOT_NULL(ptr) ((ptr) != NULL)


/*
 * Asn1Error(char* str) - configure error handler
 */
#define Asn1Error(str) fprintf(stderr,"%s", str);


/*
 * Asn1Warning(char* str) - configure warning mechanism
 * (currently never called)
 */
#define Asn1Warning(str) fprintf(stderr,"%s", str);


/*
 * configure memory scheme used by decoder to allocate memory
 * for the decoded value.
 * The Asn1Free will be called in the optionally generated
 * hierachical free routines.
 *
 */
#include "../mem.h"

#define Asn1Alloc(size)  Malloc(size)
#define Asn1Free(ptr)    Free(ptr)
#define CheckAsn1Alloc(ptr, env)\
   if ((ptr) == NULL)\
      longjmp(env, -27);


#define ENV_TYPE jmp_buf

/*
 * configure buffer routines that the encoders (write)
 * and decoders(read) use.  This config technique kind
 * of bites but is allows efficient macro calls.  The
 * Generated code & lib routines  call/use the "Buf????"
 * version of the macro - you define their meaning here.
 */
#include "exp_buf.h"

#define BUF_TYPE ExpBuf**
#define BufGetByte(b)              ExpBufGetByte(b)
#define BufGetSeg(b, lenPtr)       ExpBufGetSeg(b, lenPtr)
#define BufCopy(dst, b, len)       ExpBufCopy( dst, b, len)
#define BufSkip(b, len)            ExpBufSkip(b, len)
#define BufPeekByte(b)             ExpBufPeekByte( b)
#define BufPutByteRvs(b, byte)     ExpBufPutByteRvs( b, byte)
#define BufPutSegRvs(b, data, len) ExpBufPutSegRvs( b, data, len)
#define BufReadError(b)            ExpBufReadError(b)
#define BufWriteError(b)           ExpBufWriteError(b)

#include "print.h"  /* for printing set up */

#endif /* conditional include */
