/*
 * exp_buf.h - read/write/alloc/free routines for a simple buffer structure
 *
 *
 * NOTE: replacing the malloc and free with a allocs/frees
 *       from/to buffer pools or similar tuned/fixed size
 *       mem mgmt will improve performance.
 *
 *  You should tune the buffer management to your environment
 *  for best results
 * 
 * MS 91
 * Copyright (C) 1992 Michael Sample and the University of British Columbia
 *
 * This library is free software; you can redistribute it and/or
 * modify it provided that this copyright/license information is retained
 * in original form.
 *
 * If you modify this file, you must clearly indicate your changes.
 *
 * This source code is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef _exp_buf_h_
#define _exp_buf_h_

typedef struct ExpBuf
{
    char*          dataStart; /* points to first valid data byte */
                              /* when empty, 1 byte past blk end (rvs write)*/
    char*          dataEnd;   /* pts to first byte AFTER last valid data byte*/
    char*          curr;      /* current location to read form */
                              /* points to next byte to read */
    struct ExpBuf* next;      /* next buf (NULL if no next buffer)*/
    struct ExpBuf* prev;      /* prev buf (NULL if no prev buffer)*/
    char*          blkStart;  /* points to first byte of the blk */
    char*          blkEnd;    /* points the first byte AFTER blks last byte */
    int            readError; /* non-zero is attempt to read past end of data*/
    int            writeError;/* non-zero is attempt write fails (no mor bufs)*/
} ExpBuf;


void PutExpBufInGenBuf PROTO((ExpBuf* eb,GenBuf* gb));

/* alloc/free related protos */
void  ExpBufInit PROTO((unsigned long dataBlkSize));
ExpBuf*  ExpBufAllocBuf();
void  ExpBufFreeBuf PROTO((ExpBuf* ptr));
char* ExpBufAllocData();
void  ExpBufFreeData PROTO((char* ptr));
ExpBuf* ExpBufAllocBufAndData();
void  ExpBufFreeBufAndData PROTO(( ExpBuf* b));
void ExpBufFreeBufAndDataList PROTO(( ExpBuf* b));
void ExpBufInstallDataInBuf PROTO((ExpBuf* b, char* data, 
                                   unsigned long int len));

/* traversing related protos */
ExpBuf* ExpBufNext PROTO((ExpBuf* b));
ExpBuf* ExpBufPrev PROTO((ExpBuf* b));
ExpBuf* ExpBufListLastBuf PROTO((ExpBuf* b));
ExpBuf* ExpBufListFirstBuf PROTO((ExpBuf* b));


/* data related protos */
int ExpBufAtEod PROTO((ExpBuf* b));
int ExpBufFull PROTO((ExpBuf* b));
int ExpBufHasNoData PROTO((ExpBuf* b));
unsigned long ExpBufDataSize PROTO((ExpBuf* b));
unsigned long ExpBufDataBlkSize PROTO((ExpBuf* b));
char* ExpBufDataPtr PROTO((ExpBuf* b));


void ExpBufResetInReadMode PROTO((ExpBuf* b));
void ExpBufResetInWriteRvsMode PROTO((ExpBuf* b));

void ExpBufCopyToFile PROTO((ExpBuf* b, FILE* f));

/* 
 * standardized reading and writing routines
 * (expected by lib enc/dec routines)
 */

void          ExpBufSkip PROTO(( ExpBuf**, unsigned long len));
int           ExpBufCopy PROTO(( char* dst, ExpBuf** b, unsigned long len));
int           ExpBufPeekCopy PROTO(( char* dst, ExpBuf** b, 
                                    unsigned long len));
unsigned char ExpBufPeekByte PROTO(( ExpBuf** b));
char*         ExpBufGetSeg PROTO((ExpBuf** b, unsigned long* len));
char*         ExpBufPeekSeg PROTO((ExpBuf** b, unsigned long* len));
void          ExpBufPutSegRvs PROTO((ExpBuf** b, char* data, unsigned long len));
unsigned char ExpBufGetByte PROTO((ExpBuf** b));
void          ExpBufPutByteRvs PROTO((ExpBuf** b, unsigned char byte));

int           ExpBufReadError PROTO((ExpBuf** b));
int           ExpBufWriteError PROTO((ExpBuf** b));


#endif /* conditional include */

